/*
 *	ASP Device Driver
 *
 *	(c) Copyright 2000 The Puffin Group Inc.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *	by Helge Deller <deller@gmx.de>
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/malloc.h>
#include <asm/gsc.h>
#include <asm/led.h>

#include "busdevice.h"

#define ASP_GSC_IRQ	3		/* hardcoded interrupt for GSC */

#define ASP_HPA_HARD	0xF0800000	/* hardcoded HPA for ASP */
#define ASP_VER_OFFSET 	0x2F020		/* offset of ASP version */

#define OFFSET_ASP_LED	0x20		/* offset to HPA of the LED */ 

#undef USE_VIPER			/* VIPER - graphics-board */

static int
asp_find_irq(struct busdevice *busdev_dev, struct hp_device *dev)
{
	int irq;
	int off = ((int) dev->hpa) & 0xffff;

	/*
	** "irq" bits below are numbered relative to most significant bit.
	*/
	switch (off) {
	case 0x1000:	irq = 30; break; /* FIXME? (29,30 or 2 !)*/ /* HIL */
	case 0x2000:	irq = 25; break; /* RS232 B */	
	case 0x3000:	irq = 26; break; /* RS232 A */
	case 0x4000:	irq = 24; break; /* Centronics/Parallel Port */
	case 0x5000:	irq = 22; break; /* SCSI */
	case 0x6000:	irq = 23; break; /* LAN */

	case 0x0000:	irq = 18; break; /* Audio ??? */
	case 0x9000:	irq = 18; break; /* Audio ??? */

	case 0xF000:	irq = 17; break; /* ASP itself */
	default: 	irq = -1; break; /* unknown */
	}

	return irq;
}

int __init
asp_init_chip(struct hp_device *d, struct pa_iodc_driver *dri)
{
	struct busdevice *asp;
	struct gsc_irq gsc_irq;
	int irq, ret;

	asp = kmalloc(sizeof(struct busdevice), GFP_KERNEL);
	if(!asp)
		return -ENOMEM;

	asp->name = "ASP";
	asp->find_irq = asp_find_irq; 

	/*  The PDC-Firmware reports the ASP at 0xF082F000, 
	    but we must use a HPA of 0xF0820000 */
	asp->hpa = (char *) ASP_HPA_HARD;
	asp->version = gsc_readb(asp->hpa + ASP_VER_OFFSET);
	printk(KERN_INFO "%s version %d at 0x%p found.\n", 
		asp->name, asp->version, asp->hpa);

	/* Stop ASP hissing for a bit */
//	asp_init_irq(asp);

	/* the IRQ ASP should use */
	ret = -EBUSY;
	irq = gsc_claim_irq(&gsc_irq, ASP_GSC_IRQ);
	if (irq < 0) {
		printk(KERN_ERR __FUNCTION__ ": cannot get GSC irq\n");
		goto out;
	}

	ret = request_irq(gsc_irq.irq, busdev_barked, 0, "asp", asp);
	if (ret < 0)
		goto out;

	/* Save this for debugging later */
	asp->parent_irq = gsc_irq.irq;
	asp->eim = ((u32) gsc_irq.txn_addr) | gsc_irq.txn_data;

#ifdef USE_VIPER
	/* Tell VIPER which IRQ ASP has been assigned. */
	gsc_writel(1<<gsc_irq.irq, 0xfffbf088);
	gsc_writel(~0, 0xfffbf088);
#endif

	/* enable IRQ's for devices below ASP */
//	gsc_writel(asp->eim, asp->hpa + OFFSET_IAR);

	/* Done init'ing, register this driver */
	ret = register_busdevice( d, asp );
	if (ret)
	    goto out;
	
	/* initialize the chassis LEDs */ 
#ifdef CONFIG_CHASSIS_LCD_LED	
	register_led_driver( DISPLAY_MODEL_OLD_ASP, LED_CMD_REG_NONE, 
		    (char *) (asp->hpa + OFFSET_ASP_LED) );
#endif

	return 0;

out:
	kfree(asp);
	return ret;
}
