#ifndef BUSDEVICE_H
#define BUSDEVICE_H

#include <linux/types.h>
#include <linux/interrupt.h>
#include <asm/hardware.h>
#include <asm/gsc.h>

#define OFFSET_IRR 0x0000   /* Interrupt request register */
#define OFFSET_IMR 0x0004   /* Interrupt mask register */
#define OFFSET_IPR 0x0008   /* Interrupt pending register */
#define OFFSET_ICR 0x000C   /* Interrupt control register */
#define OFFSET_IAR 0x0010   /* Interrupt address register */


struct busdevice
{
	struct busdevice *next;
	struct hp_device *gsc;
	volatile char *hpa;
	char *name;
	int version;
	int type;
	int parent_irq;
	int eim;
	struct irq_region *busdev_region;
	spinlock_t spinlock;
	int (*find_irq) (struct busdevice *this_dev, struct hp_device *dev);
};

/* short cut to keep the compiler happy */
#define BUSDEV_DEV(x)	((struct busdevice *) (x))

int register_busdevice( struct hp_device *gsc_parent, struct busdevice *busdev_new );

/* returns a virtual irq for device at dev->hpa (works for all LASI/ASP/WAX) */
int busdevice_alloc_irq( struct hp_device *dev );

void busdev_barked(int busdev_irq, void *dev, struct pt_regs *regs);

extern unsigned int hil_base;	/* BASE of HIL-Port */
extern unsigned int hil_irq;	/* IRQ  of HIL-Port */

int lasi_init_chip(struct hp_device *d, struct pa_iodc_driver * dri);
int asp_init_chip(struct hp_device *d, struct pa_iodc_driver * dri);
int wax_init_chip(struct hp_device *d, struct pa_iodc_driver * dri);
/* int eisa_wax_init_chip(struct hp_device *d, struct pa_iodc_driver * dri); */

#endif /* BUSDEVICE_H */
