/*
 *	HIL Device Initialisation for LASI/ASP/WAX
 *
 *	(c) Copyright 2000 The Puffin Group Inc.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *	by Helge Deller <deller@gmx.de>
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/malloc.h>
#include <asm/gsc.h>

#include "busdevice.h"

unsigned int hil_base;	/* BASE of HIL-Port, will be set by WAX or ASP ! */
unsigned int hil_irq;

extern int hil_keyb_init(void);

static int __init
hil_init_chip(struct hp_device *d, struct pa_iodc_driver *dri)
{
	int	retval, irq;

	irq = busdevice_alloc_irq(d); 

	if (!irq) {
	    printk(__FILE__ ": IRQ not found for HIL at 0x%p\n", d->hpa);
	    return -ENODEV;
	}

	hil_base = (unsigned int) d->hpa;
	hil_irq  = irq;	/* store the IRQ */
	
	printk(KERN_INFO "Found HIL at 0x%x, IRQ %d\n", hil_base, hil_irq);

	/* hil_keyb_init() uses hil_base and hil_irq */
	retval = hil_keyb_init(); /* Human Interface Loop driver */

	return retval;
}


static struct pa_iodc_driver hil_drivers_for[] __initdata = {
  {HPHW_FIO, 0x0, 0x0, 0x73, 0x0, 0,
	DRIVER_CHECK_SVERSION + DRIVER_CHECK_HWTYPE,
	"HIL", "712, 715 or similiar", (void *) hil_init_chip},
  { 0 }
};

void __init 
register_hil_drivers(void)
{
	register_driver(hil_drivers_for);
}
