/*
 *	LAN Device Initialisation for LASI/ASP
 *
 *	(c) Copyright 2000 The Puffin Group Inc.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *	by Helge Deller <deller@gmx.de>
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/malloc.h>
#include <asm/gsc.h>

#include "busdevice.h"

#include <linux/netdevice.h>


extern int __init lasi_i82596_probe(struct net_device *dev);
extern int __init asp_i82596_probe(struct net_device *dev);


static int __init
lan_init_chip(struct hp_device *d, struct pa_iodc_driver *dri)
{
	struct net_device *netdevice;
	int	retval, irq;

	irq = busdevice_alloc_irq(d); 

	if (!irq) {
	    printk(KERN_ERR __FILE__ ": IRQ not found for i82596 at 0x%p\n", d->hpa);
	    return -ENODEV;
	}

	printk(KERN_INFO "Found i82596 at 0x%p, IRQ %d\n", d->hpa, irq);

	netdevice = kmalloc(sizeof(*netdevice), GFP_KERNEL);
	if (!netdevice)
	        return -ENOMEM;

	memset(netdevice, 0, sizeof(*netdevice));
	strcpy(netdevice->name,"eth0");
	netdevice->base_addr = (unsigned long) d->hpa;
	netdevice->irq = irq;
	netdevice->init =
		(d->sversion == 0x72) ? asp_i82596_probe :lasi_i82596_probe;

	retval = register_netdevice(netdevice);

	if (retval) {
		printk(KERN_WARNING __FILE__ ": register_netdevice ret'd %d\n", retval);
		kfree(netdevice);
		return -ENODEV;
	};

	return retval;
}


static struct pa_iodc_driver lan_drivers_for[] __initdata = {
  {HPHW_FIO, 0x0, 0x0, 0x8a, 0x0, 0,
	DRIVER_CHECK_SVERSION + DRIVER_CHECK_HWTYPE,
	"Lasi i82596 LAN", "712, 715 or similiar", (void *) lan_init_chip},
  {HPHW_FIO, 0x0, 0x0, 0x72, 0x0, 0,
	DRIVER_CHECK_SVERSION + DRIVER_CHECK_HWTYPE,
	"ASP i82596 LAN", "700 or similiar (old)", (void *) lan_init_chip},
  { 0 }
};

void __init 
register_lan_drivers(void)
{
	register_driver(lan_drivers_for);
}
