/*
 *	Serial Device Initialisation for LASI/ASP/WAX
 *
 *	(c) Copyright 2000 The Puffin Group Inc.
 *	(c) Copyright 2000-2001 Helge Deller <deller@gmx.de>
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/timex.h>
#include <linux/malloc.h>
#include <linux/random.h>
#include <linux/serial.h>

#include <asm/serial.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/gsc.h>

#include "busdevice.h"


static int serial_line_nr;

static int __init 
serial_init_chip(struct hp_device *d, struct pa_iodc_driver *dri)
{
	struct serial_struct *serial;
	int	retval, irq;

	irq = busdevice_alloc_irq(d); 
	if (!irq) {
#ifndef CONFIG_GSC_WAX
	    if (serial_line_nr == 1)
		printk(KERN_INFO "This kernel was compiled without WAX support, but apparently you will need to\n"
			"enable WAX support in order to use the 2nd serial port (ttyS1) on this machine.\n");
#endif
	    printk(KERN_WARNING "IRQ not found for serial device at 0x%p\n", d->hpa);
	    return -ENODEV;
	}

	serial = kmalloc(sizeof(*serial), GFP_KERNEL);
	if (!serial)
		return -ENOMEM;

	memset(serial, 0, sizeof(struct serial_struct));

	/* autoconfig() sets state->type.
	** Here we are setting the "info->type"
	*/
	serial->type = PORT_16550A;

	serial->line = serial_line_nr;
	serial->iomem_base = (void*) (((unsigned long) d->hpa) + 0x800);

#ifdef CONFIG_SERIAL_CONSOLE
	/* FIXME: inb/outb have been aliased to gsc_readb/writeb in serial.c */
	serial->port = (unsigned long) serial->iomem_base;
#endif

	serial->irq = irq; 
	serial->io_type = SERIAL_IO_GSC;	/* define access method */
	serial->flags = 0;
	serial->xmit_fifo_size = 16;
	serial->custom_divisor = 0;
	serial->baud_base = LASI_BASE_BAUD;

	retval = register_serial(serial);
	if (retval < 0) {
		printk(KERN_WARNING "%s:%s: register_serial ret'd %d\n", 
			    __FILE__, __FUNCTION__, retval);
		kfree(serial);
		return -ENODEV;
	}

	serial_line_nr++;
	return 0;
}


static struct pa_iodc_driver serial_drivers_for[] __initdata = {
  {HPHW_FIO, 0x05F, 0x0, 0x00081, 0x0, 0,		/* A-class 180 */
	DRIVER_CHECK_HVERSION + DRIVER_CHECK_HVERSION_REV + 
	DRIVER_CHECK_SVERSION + DRIVER_CHECK_HWTYPE,
	"serial device", "unknown", (void *) serial_init_chip},
  {HPHW_FIO, 0x0, 0x0, 0x8c, 0x0, 0,			/* 715/64 */
	DRIVER_CHECK_SVERSION + DRIVER_CHECK_HWTYPE,
	"serial device", "715/new or similiar", (void *) serial_init_chip},
  {HPHW_FIO, 0x0, 0x0, 0x75, 0x0, 0,			/* 715/33 */
	DRIVER_CHECK_SVERSION + DRIVER_CHECK_HWTYPE,
	"serial device", "712, 715/old or similiar", (void *) serial_init_chip},
  { 0 }
};

void __init 
probe_serial_gsc(void)
{
	register_driver(serial_drivers_for);
}
