/*
 *	WAX Device Driver
 *
 *	(c) Copyright 2000 The Puffin Group Inc.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *	by Helge Deller <deller@gmx.de>
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/gsc.h>
#include <asm/irq.h>

#include "busdevice.h"


#define WAX_GSC_IRQ	7	/* Hardcoded Interrupt for GSC */
#define WAX_GSC_NMI_IRQ	29

static struct pa_iodc_driver wax_driver_for[] = {
  	{HPHW_BA, 0x0, 0, 0x0008e, 0, 0,
       		 DRIVER_CHECK_HWTYPE + DRIVER_CHECK_SVERSION,
        	"Wax", "generic", (void *) wax_init_chip},
	{0}
};

int __init register_wax_driver(void) {
	register_driver(wax_driver_for);
}

static int wax_find_irq(struct busdevice *lasi_dev, struct hp_device *dev)
{
	int irq;
	int off = ((int) dev->hpa) & 0xffff;

	/*
	** "irq" bits below are numbered relative to most significant bit.
	*/
	switch (off) {
	case 0x0000:	irq = -1; break; /* FIXME? WAX itself */
	case 0x1000:	irq = 30; break; /* HIL */
	case 0x2000:	irq = 25; break; /* RS232 No.2 on WAX (No.1 is on Lasi) */
	default: 	irq = -1; break; /* unknown */
	}

	return irq;
}

#ifdef CONFIG_HIL
/* Reset (NMI) via HIL-Keyboard. */
static void 
wax_hil_nmi(int irq, void *handle, struct pt_regs *regs)
{
    printk(KERN_CRIT "HIL-NMI !\n");
}
#endif

static void __init
wax_init_irq(struct busdevice *wax)
{
	char *base = (char *) wax->hpa;

	/* Stop WAX barking for a bit */
	gsc_writel(0x00000000, base+OFFSET_IMR);

	/* clear pending interrupts */
	(volatile u32) gsc_readl(base+OFFSET_IRR);

	/* We're not really convinced we want to reset the onboard
         * devices. Firmware does it for us...
	 */

	/* Resets */
//	gsc_writel(0xFFFFFFFF, base+0x1000); /* HIL */
//	gsc_writel(0xFFFFFFFF, base+0x2000); /* RS232-B on Wax */
	
	/* Ok we hit it on the head with a hammer, our Dog is now
	** comatose and muzzled.  Devices will now unmask WAX
	** interrupts as they are registered as irq's in the WAX range.
	*/
}

int __init
wax_init_chip(struct hp_device *d, struct pa_iodc_driver *dri)
{
	struct busdevice *wax;
	struct gsc_irq gsc_irq;
	int irq, ret;

	wax = kmalloc(sizeof(struct busdevice), GFP_KERNEL);
	if(!wax)
		return -ENOMEM;

	wax->name = "Wax";
	wax->hpa = d->hpa;
	wax->find_irq = wax_find_irq; 

	wax->version = 0;   /* gsc_readb(wax->hpa+WAX_VER); */
	printk(KERN_INFO "%s at 0x%p found.\n", wax->name, wax->hpa);

	/* Stop wax hissing for a bit */
	wax_init_irq(wax);

	/* the IRQ wax should use */
	irq = gsc_claim_irq(&gsc_irq, WAX_GSC_IRQ);
	if (irq < 0) {
		printk(KERN_ERR __FUNCTION__ ": cannot get GSC irq\n");
		kfree(wax);
		return -EBUSY;
	}

	ret = request_irq(gsc_irq.irq, busdev_barked, 0, "wax", wax);
	if (ret < 0) {
		kfree(wax);
		return ret;
	}

	/* Save this for debugging later */
	wax->parent_irq = gsc_irq.irq;
	wax->eim = ((u32) gsc_irq.txn_addr) | gsc_irq.txn_data;

	/* enable IRQ's for devices below WAX */
//	gsc_writel(wax->eim, wax->hpa + OFFSET_IAR);

	/* Done init'ing, register this driver */
	ret = register_busdevice( d, wax );
	if (ret) 
	    kfree(wax);

	/* Register the HIL-Keyboard NMI-Handler */
#ifdef CONFIG_HIL
	request_irq( (wax->busdev_region->data.irqbase) + WAX_GSC_NMI_IRQ,
	    &wax_hil_nmi, 0, "wax_hil_nmi", NULL);
	printk("%s: HIL Keyboard-NMI registered.\n", wax->name);
#endif

#ifdef CONFIG_WAX_EISA
	register_wax_eisa_driver();
#endif

	return ret;
}

