/*
 *	WAX EISA Bus Adapter Device Driver
 *
 *	(c) Copyright 2000 Linuxcare Canada
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *	by Alex deVries <alex@linuxcare.com>
 */

#include <linux/errno.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/gsc.h>
#include <asm/irq.h>
#include <asm/pci.h>
#include <asm/processor.h>

#include "busdevice.h"

static int wax_eisa_driver_init(struct hp_device *d, struct pa_iodc_driver *dri);

static struct pa_iodc_driver wax_eisa_drivers_for[] = {
	{HPHW_BA, 0x01B, 0, 0x0008e, 0x0, 0x0,
		DRIVER_CHECK_HWTYPE + DRIVER_CHECK_SVERSION,
		"Wax EISA BA", "Wax EISA BA", (void *) wax_eisa_driver_init},
	{HPHW_BA, 0x01B, 0, 0x00076, 0x0, 0x0,
		DRIVER_CHECK_HWTYPE + DRIVER_CHECK_SVERSION,
		"Wax EISA BA", "Wax EISA BA", (void *) wax_eisa_driver_init},
	{0,0,0,0,0,0,0,
	(char *) NULL,(char *) NULL,(void *) NULL}
};

#define WAX_EISA_OUT(type, size) \
static void wax_out##type (u##size d, int addr) \
{ \
	u32 out_addr; \
	if (((addr >= 0x00080000) && (addr < 0x00100000)) || \
		((addr >= 0x00500000) && (addr < 0x03C00000))) { \
		out_addr = 0xfc000000 + ((addr & 0xfc00) >> 6) + \
			((addr & 0x03f8) << 9) + (addr & 0x0007) ; \
		gsc_write##type(d,out_addr); \
	} else { \
		printk("Wax EISA: Ack, cannot write to 0x%x\n",addr); \
	} \
} 

WAX_EISA_OUT(b,8)
WAX_EISA_OUT(w,16)
WAX_EISA_OUT(l,32)

#define WAX_EISA_IN(type, size) \
static u##size wax_in##type (int addr) \
{ \
	u32 out_addr; \
	if (((addr >= 0x00080000) && (addr < 0x00100000)) || \
		((addr >= 0x00500000) && (addr < 0x03C00000))) { \
		out_addr = 0xfc000000 + ((addr & 0xfc00) >> 6) + \
			((addr & 0x03f8) << 9) + (addr & 0x0007) ; \
		return gsc_read##type(out_addr); \
	} else { \
		printk("WAX EISA Ack, cannot read from 0x%x\n",addr); \
		return -1; \
	} \
}

WAX_EISA_IN(b,8)
WAX_EISA_IN(w,16)
WAX_EISA_IN(l,32)

static struct pci_port_ops wax_port_ops = {
	wax_inb, wax_inl, wax_inw,
	wax_outb, wax_outl, wax_outw
};

struct pci_hba_data wax_eisa_hba;

#define outsl(x) outl((x))
#define insl(x) inl((x))

/*
The following need to be created still.
memset_io
isa_memcpy_toio(a,b,c) memcpy_toio
isa_memset_io
memcpy_frioio
isa_memcpy_from_io

*/
#define isa_readl(x)	inl((x))



void __init
register_wax_eisa_driver(void)
{
        register_driver(wax_eisa_drivers_for);
}

static int
wax_eisa_driver_init(struct hp_device *d, struct pa_iodc_driver *dri) {

	int version;

	/* It sounds unusual, but because the parisc-linux PCI infrastructure
	   has been done so well, I'm just going to register the WAX EISA
	   bus adapter as a PCI device.  This is to handle the in{w|l|b}
	   and out{w|l|b} routines. */

	version = gsc_readw(d->hpa+0x7000);
	printk(KERN_INFO "Wax EISA bus adapter version 0x%x at 0x%p\n",version,d->hpa);

	pcibios_register_hba((struct pci_hba_data *) &wax_port_ops);

#if 0
	/* Turn off lock control */
	gsc_writeb(0x0,d->hpa + 0x10001); 

	/* Disable the FIFO */
	gsc_writeb(0x0,d->hpa + 0x11001);

	/* Disable multiple splits and concurrent bus accesses */
	gsc_writeb(0x0,d->hpa + 0x12001);

#endif

	EISA_bus = 1;

	return 0;

}

