/******************************************************************************/
/*                                                                            */
/* Broadcom BCM5700 Linux Network Driver, Copyright (c) 2000 Broadcom         */
/* Corporation.                                                               */
/* All rights reserved.                                                       */
/*                                                                            */
/* This program is free software; you can redistribute it and/or modify       */
/* it under the terms of the GNU General Public License as published by       */
/* the Free Software Foundation, located in the file LICENSE.                 */
/*                                                                            */
/* This file is based on skeleton.c written by Donald Becker.                 */
/*                                                                            */
/******************************************************************************/


#define SMP_CHECK
#define GLOBAL_LOCK

#ifdef NICE_SUPPORT
static const char version[] = "bcm5700.c with Broadcom NIC Extension (NICE) v1.4.5 (01/31/01)\n";
#else
static const char version[] = "bcm5700.c v1.4.5 (01/31/01)\n";
#endif

/* A few user-configurable values. */

#define MAX_UNITS 8
/* Used to pass the full-duplex flag, etc. */
int line_speed[MAX_UNITS] = {0, };
static int full_duplex[MAX_UNITS] = {1, };
static int rx_flow_control[MAX_UNITS] = {0, };
static int tx_flow_control[MAX_UNITS] = {0, };
static int auto_flow_control[MAX_UNITS] = {0, };
static int mtu[MAX_UNITS] = {0, };			/* Jumbo MTU for interfaces. */

/* Operational parameters that usually are not changed. */
/* Time in jiffies before concluding the transmitter is hung. */
#define TX_TIMEOUT  (2*HZ)

#include <linux/config.h>
#if defined(CONFIG_SMP) && ! defined(__SMP__)
#define __SMP__
#endif
#if defined(CONFIG_MODVERSIONS) && defined(MODULE) && ! defined(MODVERSIONS)
#define MODVERSIONS
#endif

#include <linux/version.h>
#ifdef MODULE
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <net/ip.h>
#include <asm/processor.h>		/* Processor type for cache alignment. */
#include <asm/bitops.h>
#include <asm/io.h>
#include <asm/unaligned.h>
#include <linux/delay.h>

#include "lm.h"
#include "queue.h"
#include "tigon3.h"
#include "mm.h"

#if (LINUX_VERSION_CODE < 0x02030e)
#define net_device device
#endif

#if (LINUX_VERSION_CODE < 0x02032b)
#define dev_kfree_skb_irq(skb)  dev_free_skb(skb)
#define netif_wake_queue(dev)	clear_bit(0, &dev->tbusy); mark_bh(NET_BH)
#define netif_stop_queue(dev)	set_bit(0, &dev->tbusy)

static inline void netif_start_queue(struct net_device *dev)
{
	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;
}

#define netif_queue_stopped(dev)	dev->tbusy
#define netif_running(dev)		dev->start
#endif

#if (LINUX_VERSION_CODE < 0x02032a)
typedef u32 dma_addr_t;

static inline void *pci_alloc_consistent(struct pci_dev *pdev, size_t size,
					 dma_addr_t *dma_handle)
{
	void *virt_ptr;

	virt_ptr = kmalloc(size, GFP_KERNEL);
	*dma_handle = virt_to_bus(virt_ptr);
	return virt_ptr;
}
#define pci_free_consistent(cookie, size, ptr, dma_ptr)	kfree(ptr)
#define pci_map_single(cookie, address, size, dir)	virt_to_bus(address)
#define pci_unmap_single(cookie, address, size, dir)
#endif

static inline void set_64bit_addr(LM_PHYSICAL_ADDRESS *paddr, dma_addr_t addr)
{
	unsigned long baddr = (unsigned long) addr;
#if (BITS_PER_LONG == 64)
	paddr->Low = baddr & 0xffffffff;
	paddr->High = baddr >> 32;
#else
	paddr->Low = baddr;
	paddr->High = 0;
#endif
}

/*
 * Broadcom NIC Extension support
 * -ffan
 */
#ifdef NICE_SUPPORT
#include "nicext.h"

typedef struct {
	ushort  tag;
	ushort  signature;
} vlan_tag_t;

#endif /* NICE_SUPPORT */

DECLARE_QUEUE_TYPE(UM_RX_PACKET_Q, MAX_RX_PACKET_DESC_COUNT+1);

#define MAX_MEM 16

typedef struct _UM_DEVICE_BLOCK {
	LM_DEVICE_BLOCK lm_dev;
	struct net_device *dev;
	struct pci_dev *pdev;
	struct net_device *next_module;
	int index;
	int delayed_link_ind;	/* indicate initial status after a delay */
	int tx_full;
	UM_RX_PACKET_Q rx_out_of_buf_q;
	int mem_list_num;
	void *mem_list[MAX_MEM];
	dma_addr_t dma_list[MAX_MEM];
	int mem_size_list[MAX_MEM];
	struct timer_list timer;
	long tx_lock[MAX_TX_LOCK];
	long rx_lock;
	int interrupt;
	spinlock_t global_lock;
	spinlock_t undi_lock;
	long undi_flags;
#if LINUX_VERSION_CODE > 0x20139
	struct net_device_stats stats;
#else
	struct enet_statistics stats;
#endif
#ifdef NICE_SUPPORT
	void (*nice_rx)( struct sk_buff*, void* );
	void* nice_ctx;
#endif /* NICE_SUPPORT */
} UM_DEVICE_BLOCK, *PUM_DEVICE_BLOCK;

typedef struct _UM_PACKET {
	LM_PACKET lm_packet;
	struct sk_buff *skbuff;
	dma_addr_t mapping;
	LM_FRAG_LIST frag_list;
} UM_PACKET, *PUM_PACKET;

int MM_Packet_Desc_Size = sizeof(UM_PACKET);

/* Kernel compatibility defines, some common to David Hind's PCMCIA package.
   This is only in the support-all-kernels source code. */

#if defined(MODULE) && LINUX_VERSION_CODE > 0x20115
MODULE_AUTHOR("Michael Chan <mchan@broadcom.com>");
MODULE_DESCRIPTION("BCM5700 Driver");
MODULE_PARM(debug, "i");
MODULE_PARM(line_speed, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(full_duplex, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(rx_flow_control, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(tx_flow_control, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(auto_flow_control, "1-" __MODULE_STRING(MAX_UNITS) "i");
MODULE_PARM(mtu, "1-" __MODULE_STRING(MAX_UNITS) "i");
#endif

#define RUN_AT(x) (jiffies + (x))

#if (LINUX_VERSION_CODE >= 0x20100)
char kernel_version[] = UTS_RELEASE;
#endif

#if LINUX_VERSION_CODE < 0x20123
#define hard_smp_processor_id() smp_processor_id()
#define test_and_set_bit(val, addr) set_bit(val, addr)
#define le16_to_cpu(val) (val)
#define le32_to_cpu(val) (val)
#define cpu_to_le32(val) (val)
#endif
#if LINUX_VERSION_CODE <= 0x20139
#define	net_device_stats enet_statistics
#else
#define NETSTATS_VER2
#endif
#if LINUX_VERSION_CODE < 0x20155
/* Grrrr, the PCI code changed, but did not consider CardBus... */
#include <linux/bios32.h>
#define PCI_SUPPORT_VER1
#else
#define PCI_SUPPORT_VER2
#endif
#if LINUX_VERSION_CODE < 0x20159
#define dev_free_skb(skb) dev_kfree_skb(skb, FREE_WRITE)
#else
#define dev_free_skb(skb) dev_kfree_skb(skb)
#endif
#if ! defined(CAP_NET_ADMIN)
#define capable(CAP_XXX) (suser())
#endif

#define tigon3_debug debug
#ifdef TIGON3_DEBUG
static int tigon3_debug = TIGON3_DEBUG;
#else
static int tigon3_debug = 1;
#endif

#if (LINUX_VERSION_CODE >= 0x10344)
#define NEW_MULTICAST
#endif

#ifdef SA_SHIRQ
#define IRQ(irq, dev_id, pt_regs) (irq, dev_id, pt_regs)
#else
#define IRQ(irq, dev_id, pt_regs) (irq, pt_regs)
#endif

/* This my implementation of shared IRQs, now only used for 1.2.13. */
#ifdef HAVE_SHARED_IRQ
#define USE_SHARED_IRQ
#include <linux/shared_irq.h>
#endif

#if 0
static struct net_device *bcm5700_probe1(struct pci_dev *pdev,
	struct net_device *dev, int chip_idx, int board_idx);

static int bcm5700_open(struct net_device *dev);
static void bcm5700_timer(unsigned long data);
static void bcm5700_tx_timeout(struct net_device *dev);
static int bcm5700_start_xmit(struct sk_buff *skb, struct net_device *dev);
static void bcm5700_interrupt IRQ(int irq, void *dev_instance, struct pt_regs *regs);
static int bcm5700_close(struct net_device *dev);
#else
struct net_device *bcm5700_probe1(struct pci_dev *pdev,
	struct net_device *dev, int chip_idx, int board_idx);
int bcm5700_open(struct net_device *dev);
void bcm5700_timer(unsigned long data);
void bcm5700_tx_timeout(struct net_device *dev);
int bcm5700_start_xmit(struct sk_buff *skb, struct net_device *dev);
void bcm5700_interrupt IRQ(int irq, void *dev_instance, struct pt_regs *regs);
int bcm5700_close(struct net_device *dev);
#endif
static struct net_device_stats *bcm5700_get_stats(struct net_device *dev);
#ifdef HAVE_PRIVATE_IOCTL
static int private_ioctl(struct net_device *dev, struct ifreq *rq, int cmd);
#endif
#ifdef NEW_MULTICAST
static void set_rx_mode(struct net_device *dev);
#else
static void set_rx_mode(struct net_device *dev, int num_addrs, void *addrs);
#endif
static int bcm5700_set_mac_addr(struct net_device *dev, void *p);
void replenish_rx_buffers(PUM_DEVICE_BLOCK pUmDevice);

/* This table drives the PCI probe routines.  It's mostly boilerplate in all
   of the drivers, and will likely be provided by some future kernel.
   Note the matching code -- the first table entry matchs all 56** cards but
   second only the 1234 card.
*/
enum pci_flags_bit {
	PCI_USES_IO=1, PCI_USES_MEM=2, PCI_USES_MASTER=4,
	PCI_ADDR0=0x10<<0, PCI_ADDR1=0x10<<1, PCI_ADDR2=0x10<<2, PCI_ADDR3=0x10<<3,
};
#define PCI_ADDR0_IO (PCI_USES_IO|PCI_ADDR0)

struct pci_id_info {
	const char *name;
	u16	vendor_id, device_id, device_id_mask, flags;
	int io_size, min_latency;
	struct net_device *(*probe1)(struct pci_dev *pdev, struct net_device *dev,
							 int chip_idx, int fnd_cnt);
};

static struct pci_id_info pci_tbl[] = {
#if 0
  { "Broadcom BCM5700",
	0x12ae, 0x0003, 0xffff, PCI_ADDR0_IO, 128, 32, bcm5700_probe1 },
#endif
  { "Broadcom BCM5700",
	0x14e4, 0x1644, 0xffff, PCI_ADDR0_IO, 128, 32, bcm5700_probe1 },
  {0},
};

/* A list of all installed Tulip devices. */
static struct net_device *root_tigon3_dev = NULL;

int bcm5700_probe(struct net_device *dev)
{
	PLM_DEVICE_BLOCK pDevice;
	int cards_found = 0;
	struct pci_dev *pdev = 0;

	if ( ! pci_present())
		return -ENODEV;

	while ((pdev = pci_find_class(PCI_CLASS_NETWORK_ETHERNET << 8, pdev))) {
		int chip_idx;

		for (chip_idx = 0; pci_tbl[chip_idx].vendor_id; chip_idx++)
			if (pdev->vendor == pci_tbl[chip_idx].vendor_id
				&& (pdev->device & pci_tbl[chip_idx].device_id_mask) ==
				pci_tbl[chip_idx].device_id)
				break;
		if (pci_tbl[chip_idx].vendor_id == 0)
			continue;


		dev = pci_tbl[chip_idx].probe1(pdev, dev, chip_idx, cards_found);
		pDevice = (PLM_DEVICE_BLOCK) dev->priv;
		if (pDevice == 0)
			continue;

		if (debug > 2)
			printk(KERN_INFO "Found %s at PCI Mem. address %#lx.\n",
				   pci_tbl[chip_idx].name, pDevice->MemBaseLow);

		/* Get and check the bus-master and latency values. */
		if (dev) {
			u8 pci_latency;
			pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &pci_latency);
			if (pci_latency < 10) {
				printk(KERN_INFO "  PCI latency timer (CFLT) is "
					   "unreasonably low at %d.  Setting to 64 clocks.\n",
					   pci_latency);
				pci_write_config_byte(pdev, PCI_LATENCY_TIMER, 64);
			}
		}
		dev = 0;
		((PUM_DEVICE_BLOCK) pDevice)->index = cards_found;
		cards_found++;
	}

	return cards_found ? 0 : -ENODEV;
}

#if 0
static struct net_device *bcm5700_probe1(struct pci_dev *pdev,
								   struct net_device *dev,
								   int chip_idx, int board_idx)
#else
struct net_device *bcm5700_probe1(struct pci_dev *pdev,
								   struct net_device *dev, 
								   int chip_idx, int board_idx)
#endif
{
	PLM_DEVICE_BLOCK pDevice;
	PUM_DEVICE_BLOCK pUmDevice;
	int i;

	dev = init_etherdev(dev, 0);

	printk(KERN_INFO "Broadcom Gigabit Ethernet Driver %s", version);

	dev->priv = kmalloc(sizeof(UM_DEVICE_BLOCK), GFP_KERNEL);
	if (dev->priv == 0) {
		printk(KERN_DEBUG "Cannot allocate private structure\n");
		return dev;
	}
	memset(dev->priv, 0, sizeof(UM_DEVICE_BLOCK));

	pUmDevice = (PUM_DEVICE_BLOCK) dev->priv;

	spin_lock_init(&pUmDevice->global_lock);
	spin_lock_init(&pUmDevice->undi_lock);

	pUmDevice->dev = dev;
	pUmDevice->pdev = pdev;
	pUmDevice->next_module = root_tigon3_dev;
	root_tigon3_dev = dev;

	pUmDevice->mem_list_num = 0;

	pDevice = (PLM_DEVICE_BLOCK) pUmDevice;

	if (mtu[board_idx] > 1500) {
		if (mtu[board_idx] > (9 * 1024))
			dev->mtu = (9 * 1024);
		else
			dev->mtu = mtu[board_idx];
	}

	if (LM_GetAdapterInfo(pDevice) != LM_STATUS_SUCCESS) {
		printk(KERN_DEBUG "Get Adapter info failed\n");
		return dev;
	}

	dev->base_addr = pDevice->MemBaseLow;
	dev->irq = pDevice->Irq = pdev->irq;
	/* The Tigon3-specific entries in the device structure. */
	dev->open = &bcm5700_open;
	dev->hard_start_xmit = &bcm5700_start_xmit;
	dev->stop = &bcm5700_close;
	dev->get_stats = &bcm5700_get_stats;
#ifdef HAVE_PRIVATE_IOCTL
	dev->do_ioctl = &private_ioctl;
#endif
#ifdef HAVE_MULTICAST
	dev->set_multicast_list = &set_rx_mode;
#endif
	dev->set_mac_address = &bcm5700_set_mac_addr;
	memcpy(dev->dev_addr, pDevice->NodeAddress, 6);
	printk(KERN_INFO "%s: %s found at mem %lx, IRQ %d, ",
		dev->name, pci_tbl[chip_idx].name, dev->base_addr, dev->irq);
	printk("node addr ");
	for (i = 0; i < 6; i++) {
		printk("%2.2x", dev->dev_addr[i]);
	}
	printk("\n");
	return dev;
}


//static int
int
bcm5700_open(struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	int next_tick = 6*HZ;

	MOD_INC_USE_COUNT;
	pDevice->TxPacketDescCnt = 16;
	if (LM_InitializeAdapter(pDevice) != LM_STATUS_SUCCESS)
		return -EAGAIN;

	if (memcmp(dev->dev_addr, pDevice->NodeAddress, 6)) {
		LM_SetMacAddress(pDevice, dev->dev_addr);
	}

	if (tigon3_debug > 1)
		printk(KERN_DEBUG "%s: tigon3_open() irq %d.\n", dev->name, dev->irq);

	printk(KERN_INFO "%s: ", dev->name);
	if ((pDevice->PhyId & PHY_ID_MASK) == PHY_BCM5400_PHY_ID)
		printk("Broadcom BCM5400 ");
	else if ((pDevice->PhyId & PHY_ID_MASK) == PHY_BCM5401_PHY_ID)
		printk("Broadcom BCM5401 ");
	else
		printk("Unknown ");
	printk("MII transceiver found\n");

	QQ_InitQueue(&pUmDevice->rx_out_of_buf_q.Container,
        MAX_RX_PACKET_DESC_COUNT);
	netif_start_queue(dev);

	init_timer(&pUmDevice->timer);
	pUmDevice->timer.expires = RUN_AT(next_tick);
	pUmDevice->timer.data = (unsigned long)dev;
	pUmDevice->timer.function = &bcm5700_timer;
	add_timer(&pUmDevice->timer);

#ifdef SA_SHIRQ
	if (request_irq(dev->irq, &bcm5700_interrupt, SA_SHIRQ, dev->name, dev)) {
		return -EAGAIN;
	}
#else
	if (irq2dev_map[dev->irq] != NULL
		|| (irq2dev_map[dev->irq] = dev) == NULL
		|| dev->irq == 0
		|| request_irq(dev->irq, &bcm5700_interrupt, 0, dev->name)) {
		return -EAGAIN;
	}
#endif
	LM_EnableInterrupt(pDevice);
	pUmDevice->delayed_link_ind = 1;
//        MM_IndicateStatus(pDevice, pDevice->LinkStatus);

	return 0;
}


//static void 
void
bcm5700_timer(unsigned long data)
{
	struct net_device *dev = (struct net_device *)data;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;

	if (pUmDevice->delayed_link_ind)
        MM_IndicateStatus(pDevice, pDevice->LinkStatus);

	if (QQ_GetEntryCnt(&pDevice->TxPacketFreeQ.Container) !=
		pDevice->TxPacketDescCnt) {

		if ((jiffies - dev->trans_start) > TX_TIMEOUT) {
			if (pDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED) {
				/* This will generate an interrupt */
				REG_WR(pDevice, Grc.LocalCtrl, pDevice->GrcLocalCtrl |
					GRC_MISC_LOCAL_CTRL_SET_INT);
			}
			else {
				printk(KERN_WARNING "%s: Tx hung\n", dev->name);
				bcm5700_tx_timeout(dev);
			}
		}
	}
	if (QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container) >=
		pDevice->RxPacketDescCnt) {

		/* This will generate an interrupt */
		pDevice->pStatusBlkVirt->Status |= STATUS_BLOCK_UPDATED;
		REG_WR(pDevice, Grc.LocalCtrl, pDevice->GrcLocalCtrl |
			GRC_MISC_LOCAL_CTRL_SET_INT);
	}
	pUmDevice->timer.expires = RUN_AT(2*HZ);
	add_timer(&pUmDevice->timer);
}

//static void 
void
bcm5700_tx_timeout(struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;

	LM_DisableInterrupt(pDevice);
	LM_ResetAdapter(pDevice);	
	if (memcmp(dev->dev_addr, pDevice->NodeAddress, 6)) {
		LM_SetMacAddress(pDevice, dev->dev_addr);
	}
	LM_EnableInterrupt(pDevice);
}

//static int
int
bcm5700_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	PLM_FRAG_LIST pfrag_list;
	long flags;
#ifdef NICE_SUPPORT
	vlan_tag_t *vlan_tag;
#endif

#ifdef GLOBAL_LOCK
	spin_lock_irqsave(&pUmDevice->global_lock, flags);
	netif_stop_queue(dev);
#else
	save_flags(flags);
	cli();
#if (LINUX_VERSION_CODE < 0x02032b)
	if (test_and_set_bit(0, &dev->tbusy)) {
		if (jiffies - dev->trans_start < TX_TIMEOUT) {
			restore_flags(flags);
			return 1;
		}
//		bcm5700_tx_timeout(dev);
		restore_flags(flags);
		return 1;
	}
#else
	netif_stop_queue(dev);
#endif
#endif
	MM_ACQUIRE_TX_LOCK(pDevice, 0, FreeQ);
	pPacket = (PLM_PACKET)
		QQ_PopHead(&pDevice->TxPacketFreeQ.Container);
	MM_RELEASE_TX_LOCK(pDevice, 0, FreeQ);
	if (pPacket == 0) {
#ifdef GLOBAL_LOCK
		spin_unlock_irqrestore(&pUmDevice->global_lock, flags);
#else
		restore_flags(flags);
#endif
		return 1;
	}
	pUmPacket = (PUM_PACKET) pPacket;
	pUmPacket->skbuff = skb;
	pPacket->Flags = 0;
	pPacket->u.Tx.TaskOffload = LM_TASK_OFFLOAD_NONE;
	pfrag_list = &pUmPacket->frag_list;
	pPacket->u.Tx.pFraglist = pfrag_list;
	pfrag_list->FragCount = pPacket->u.Tx.FragCount = 1;
	pfrag_list->TotalSize = pPacket->PacketSize = skb->len;
	pfrag_list->Fragments[0].FragSize = skb->len;
	pUmPacket->mapping = pci_map_single(pUmDevice->pdev, skb->data,
				skb->len, PCI_DMA_TODEVICE);
	set_64bit_addr(&pfrag_list->Fragments[0].FragBuf, pUmPacket->mapping);

#ifdef NICE_SUPPORT
	vlan_tag = (vlan_tag_t *) &skb->cb[0];
	if (vlan_tag->signature == 0x5555) {
		pPacket->VlanTag = vlan_tag->tag;
		pPacket->Flags |= LM_VALID_VLAN_TAG;
		vlan_tag->signature = 0;
	}
#endif
	LM_SendPacket(pDevice, pPacket);
	if (QQ_GetEntryCnt(&pDevice->TxPacketFreeQ.Container) != 0) {
		netif_wake_queue(dev);
	}
	else {
		pUmDevice->tx_full = 1;
	}
	dev->trans_start = jiffies;
#ifdef GLOBAL_LOCK
	spin_unlock_irqrestore(&pUmDevice->global_lock, flags);
#else
	restore_flags(flags);
#endif
	return 0;
}

//static void
void
bcm5700_interrupt IRQ(int irq, void *dev_instance, struct pt_regs *regs)
{
#ifdef SA_SHIRQ		/* Use the now-standard shared IRQ implementation. */
	struct net_device *dev = (struct net_device *)dev_instance;
#else
	struct net_device *dev = (struct net_device *)(irq2dev_map[irq]);
#endif
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;

#ifdef GLOBAL_LOCK
	spin_lock(&pUmDevice->global_lock);
#else
#if defined(__i386__) && defined(SMP_CHECK)
	if (test_and_set_bit(0, (void*)&pUmDevice->interrupt)) {
		printk(KERN_ERR "%s: Duplicate entry of the interrupt handler by "
			   "processor %d.\n",
			   dev->name, hard_smp_processor_id());
		pUmDevice->interrupt = 0;
		return;
	}
#else
	if (pUmDevice->interrupt) {
		printk(KERN_ERR "%s: Re-entering the interrupt handler.\n", dev->name);
		return;
	}
	pUmDevice->interrupt = 1;
#endif
#endif

	while (pDevice->pStatusBlkVirt->Status & STATUS_BLOCK_UPDATED) {
        ulong dummy;

		pDevice->pMemView->Mailbox.Interrupt[0].Low = 1;
   		pDevice->pStatusBlkVirt->Status &= ~STATUS_BLOCK_UPDATED;
		LM_ServiceInterrupts(pDevice);
		pDevice->pMemView->Mailbox.Interrupt[0].Low = 0;
		dummy = pDevice->pMemView->Mailbox.Interrupt[0].Low;
	}
	if (QQ_GetEntryCnt(&pUmDevice->rx_out_of_buf_q.Container)) {
		replenish_rx_buffers(pUmDevice);
	}

#ifdef GLOBAL_LOCK
	spin_unlock(&pUmDevice->global_lock);
#else
#if defined(__i386__)
	clear_bit(0, (void*)&pUmDevice->interrupt);
#else
	pUmDevice->interrupt = 0;
#endif
#endif
	return;
}

//static int
int
bcm5700_close(struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	int i;

#if (LINUX_VERSION_CODE < 0x02032b)
	dev->start = 0;
#endif
	netif_stop_queue(dev);

	if (tigon3_debug > 1)
		printk(KERN_DEBUG "%s: Shutting down Tigon3\n",
			   dev->name);

	LM_DisableInterrupt(pDevice);
	LM_Halt(pDevice);
	pDevice->InitDone = 0;
	del_timer(&pUmDevice->timer);

	free_irq(dev->irq, dev);
	for (i = 0; i < pUmDevice->mem_list_num; i++) {
		if (pUmDevice->dma_list[i] == 0) {
			kfree(pUmDevice->mem_list[i]);
		}
		else {
			pci_free_consistent(pUmDevice->pdev,
				(size_t) pUmDevice->mem_size_list[i],
				pUmDevice->mem_list[i],
				pUmDevice->dma_list[i]);
		}
	}
	pUmDevice->mem_list_num = 0;
	MOD_DEC_USE_COUNT;

	return 0;
}

static struct net_device_stats *
bcm5700_get_stats(struct net_device *dev)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	PT3_STATS_BLOCK pStats = (PT3_STATS_BLOCK) pDevice->pStatsBlkVirt;
	struct net_device_stats *p_netstats = &pUmDevice->stats;

	if (pStats == 0)
		return p_netstats;

	/* Get stats from LM */
	p_netstats->rx_packets = pStats->ifHCInUcastPkts.Low +
			pStats->ifHCInMulticastPkts.Low +
			pStats->ifHCInBroadcastPkts.Low;
	p_netstats->tx_packets = pStats->COSIfHCOutPkts[0].Low;
	p_netstats->rx_bytes = pStats->ifHCInOctets.Low;
	p_netstats->tx_bytes = pStats->ifHCOutOctets.Low;
	p_netstats->rx_errors = pStats->dot3StatsFCSErrors.Low +
			pStats->dot3StatsAlignmentErrors.Low +
			pStats->dot3StatsFramesTooLong.Low +
			pStats->etherStatsJabbers.Low +
			pStats->etherStatsUndersizePkts.Low +
			pStats->ifInErrors.Low;
	p_netstats->tx_errors = pStats->dot3StatsInternalMacTransmitErrors.Low +
			pStats->dot3StatsCarrierSenseErrors.Low +

			pStats->ifOutDiscards.Low +
			pStats->ifOutErrors.Low;
	p_netstats->multicast = pStats->ifHCInMulticastPkts.Low;
	p_netstats->collisions = pStats->etherStatsCollisions.Low;
	p_netstats->rx_length_errors = pStats->dot3StatsFramesTooLong.Low +
			pStats->etherStatsUndersizePkts.Low;
	p_netstats->rx_over_errors = pStats->nicNoMoreRxBDs.Low;
	p_netstats->rx_crc_errors = pStats->dot3StatsFCSErrors.Low;
	p_netstats->rx_frame_errors = pStats->dot3StatsAlignmentErrors.Low;
	p_netstats->tx_aborted_errors = pStats->ifOutDiscards.Low;
	p_netstats->tx_carrier_errors = pStats->dot3StatsCarrierSenseErrors.Low;

	return p_netstats;
}

#ifdef HAVE_PRIVATE_IOCTL
/* Provide ioctl() calls to examine the MII xcvr state. */
static int private_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	u16 *data = (u16 *)&rq->ifr_data;
	u32 value;
	unsigned long flags;

	switch(cmd) {
	case SIOCDEVPRIVATE:		/* Get the address of the PHY in use. */
		data[0] = pDevice->PhyAddr;
	case SIOCDEVPRIVATE+1:		/* Read the specified MII register. */
#ifdef GLOBAL_LOCK
		spin_lock_irqsave(&pUmDevice->global_lock, flags);
#endif
		LM_ReadPhy(pDevice, data[1] & 0x1f, (LM_UINT32 *) &value);
#ifdef GLOBAL_LOCK
		spin_unlock_irqrestore(&pUmDevice->global_lock, flags);
#endif
		data[3] = value & 0xffff;
		return 0;
	case SIOCDEVPRIVATE+2:		/* Write the specified MII register */
		if (!capable(CAP_NET_ADMIN))
			return -EPERM;
#ifdef GLOBAL_LOCK
		spin_lock_irqsave(&pUmDevice->global_lock, flags);
#endif
		LM_WritePhy(pDevice, data[1] & 0x1f, data[2]);
#ifdef GLOBAL_LOCK
		spin_unlock_irqrestore(&pUmDevice->global_lock, flags);
#endif
		return 0;
#ifdef NICE_SUPPORT
	case SIOCNICE:
		{
			struct nice_req* nrq;
			nrq = (struct nice_req*)&rq->ifr_ifru;
			if( nrq->cmd == NICE_CMD_QUERY_SUPPORT ) {
				nrq->nrq_magic = NICE_DEVICE_MAGIC;
				nrq->nrq_support_rx = 1;
				return 0;
			}
			else if( nrq->cmd == NICE_CMD_SET_RX ) {
				pUmDevice->nice_rx = nrq->nrq_rx;
				pUmDevice->nice_ctx = nrq->nrq_ctx;
				return 0;
			}
			else if( nrq->cmd == NICE_CMD_GET_RX ) {
				nrq->nrq_rx = pUmDevice->nice_rx;
				nrq->nrq_ctx = pUmDevice->nice_ctx;
				return 0;
			}
			break;
		}
#endif /* NICE_SUPPORT */
	default:
		return -EOPNOTSUPP;
	}
	return -EOPNOTSUPP;
}
#endif  /* HAVE_PRIVATE_IOCTL */

#ifdef NEW_MULTICAST
static void set_rx_mode(struct net_device *dev)
#else
static void set_rx_mode(struct net_device *dev, int num_addrs, void *addrs)
#endif
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK)dev->priv;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	int i;
	struct dev_mc_list *mclist;

	for (i = 0, mclist = dev->mc_list; mclist && i < dev->mc_count;
			 i++, mclist = mclist->next) {
		LM_MulticastAdd(pDevice, (PLM_UINT8) &mclist->dmi_addr);
	}
}

/*
 * Set the hardware MAC address.
 */
static int bcm5700_set_mac_addr(struct net_device *dev, void *p)
{
	struct sockaddr *addr=p;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) dev->priv;

	if (netif_running(dev))
		return -EBUSY;
	memcpy(dev->dev_addr, addr->sa_data,dev->addr_len);
	LM_SetMacAddress(pDevice, dev->dev_addr);
	return 0;
}


#ifdef MODULE
int init_module(void)
{
	return bcm5700_probe(NULL);
}

void cleanup_module(void)
{
	struct net_device *next_dev;
	PUM_DEVICE_BLOCK pUmDevice;

	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */
	while (root_tigon3_dev) {
		pUmDevice = (PUM_DEVICE_BLOCK)root_tigon3_dev->priv;
		next_dev = pUmDevice->next_module;
		unregister_netdev(root_tigon3_dev);
		kfree(root_tigon3_dev);
		kfree(pUmDevice);
		root_tigon3_dev = next_dev;
	}
}

#endif  /* MODULE */
/*
 * Middle Module
 *
 */


#define STD_BUF_SZ	1536
#define MIN_BUF_SZ	128

LM_STATUS
MM_ReadConfig32(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT32 *pValue32)
{
	UM_DEVICE_BLOCK *pUmDevice;

	pUmDevice = (UM_DEVICE_BLOCK *) pDevice;
	pci_read_config_dword(pUmDevice->pdev, Offset, (u32 *) pValue32);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_WriteConfig32(PLM_DEVICE_BLOCK pDevice, LM_UINT32 Offset,
	LM_UINT32 Value32)
{
	UM_DEVICE_BLOCK *pUmDevice;

	pUmDevice = (UM_DEVICE_BLOCK *) pDevice;
	pci_write_config_dword(pUmDevice->pdev, Offset, Value32);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_AllocateSharedMemory(PLM_DEVICE_BLOCK pDevice, LM_UINT32 BlockSize,
	PLM_VOID *pMemoryBlockVirt, PLM_PHYSICAL_ADDRESS pMemoryBlockPhy,
	LM_BOOL Cached)
{
	PLM_VOID pvirt;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	dma_addr_t mapping;

	pvirt = pci_alloc_consistent(pUmDevice->pdev, BlockSize,
					       &mapping);
	if (!pvirt)
		return LM_STATUS_FAILURE;
	pUmDevice->mem_list[pUmDevice->mem_list_num] = pvirt;
	pUmDevice->dma_list[pUmDevice->mem_list_num] = mapping;
	pUmDevice->mem_size_list[pUmDevice->mem_list_num++] = BlockSize;
	memset(pvirt, 0, BlockSize);
	*pMemoryBlockVirt = (PLM_VOID) pvirt;
	set_64bit_addr(pMemoryBlockPhy, mapping);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_AllocateMemory(PLM_DEVICE_BLOCK pDevice, LM_UINT32 BlockSize,
	PLM_VOID *pMemoryBlockVirt)
{
	PLM_VOID pvirt;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;

	pvirt = kmalloc(BlockSize, GFP_KERNEL);
	if (!pvirt)
		return LM_STATUS_FAILURE;
	pUmDevice->mem_list[pUmDevice->mem_list_num] = pvirt;
	pUmDevice->dma_list[pUmDevice->mem_list_num] = 0;
	pUmDevice->mem_size_list[pUmDevice->mem_list_num++] = BlockSize;
	memset(pvirt, 0, BlockSize);
	*pMemoryBlockVirt = pvirt;
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_MapMemBase(PLM_DEVICE_BLOCK pDevice)
{
	pDevice->pMappedMemBase = ioremap_nocache(pDevice->MemBaseLow,
		pDevice->MemBaseSize);
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_InitializeUmPackets(PLM_DEVICE_BLOCK pDevice)
{
	int i;
	struct sk_buff *skb;
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	PUM_PACKET pUmPacket;
	PLM_PACKET pPacket;

	for (i = 0; i < pDevice->RxPacketDescCnt; i++) {
		pPacket = QQ_PopHead(&pDevice->RxPacketFreeQ.Container);
		pUmPacket = (PUM_PACKET) pPacket;
		if (pPacket == 0) {
			printk(KERN_DEBUG "Bad RxPacketFreeQ\n");
		}
		skb = dev_alloc_skb(pPacket->u.Rx.RxBufferSize);
		if (skb == 0) {
			printk(KERN_DEBUG "Null skb\n");
		}
		pUmPacket->skbuff = skb;
		skb->dev = pUmDevice->dev;
//		skb->protocol = eth_type_trans(skb, skb->dev);
		skb_reserve(skb, 2);
		pUmPacket->mapping = pci_map_single(pUmDevice->pdev,
					skb->tail,
					pPacket->u.Rx.RxBufferSize,
					PCI_DMA_FROMDEVICE);
		pPacket->u.Rx.pRxBufferVirt = skb->tail;
		set_64bit_addr(&pPacket->u.Rx.RxBufferPhy, pUmPacket->mapping);
		QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
	}
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_GetConfig(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	int index;

	pDevice->TaskToOffload |= LM_TASK_OFFLOAD_NONE;
//	pDevice->StatsCoalescingTicks = 0;

	index = pUmDevice->index;
	if (line_speed[index] != 0) {
/*		if (line_speed[index] == 1000) {
			if (full_duplex[index]) {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_1000MBPS_FULL_DUPLEX;
			}
			else {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_1000MBPS;
			}
		}*/
		if (line_speed[index] == 100) {
			if (full_duplex[index]) {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_100MBPS_FULL_DUPLEX;
			}
			else {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_100MBPS;
			}
		}
		if (line_speed[index] == 10) {
			if (full_duplex[index]) {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_10MBPS_FULL_DUPLEX;
			}
			else {
				pDevice->RequestedMediaType =
					LM_REQUESTED_MEDIA_TYPE_UTP_10MBPS;
			}
		}

		pDevice->DisableAutoNeg = TRUE;
	}
	pDevice->FlowControlCap = 0;
	if (rx_flow_control[index] != 0) {
		pDevice->FlowControlCap |= LM_FLOW_CONTROL_RECEIVE_PAUSE;
	}
	if (tx_flow_control[index] != 0) {
		pDevice->FlowControlCap |= LM_FLOW_CONTROL_TRANSMIT_PAUSE;
	}
	if ((auto_flow_control[index] != 0) && (line_speed[index] == 0)) {
		pDevice->FlowControlCap |= LM_FLOW_CONTROL_AUTO_PAUSE;
		if ((tx_flow_control[index] == 0) && (rx_flow_control[index] == 0)) {
			pDevice->FlowControlCap |=
					LM_FLOW_CONTROL_TRANSMIT_PAUSE |
					LM_FLOW_CONTROL_RECEIVE_PAUSE;
		}
	}

	if (pUmDevice->dev->mtu > 1500) {
		pDevice->RxMtu = pUmDevice->dev->mtu;
		pDevice->RxJumboDescCnt = DEFAULT_JUMBO_RCV_DESC_COUNT;
	}
	else {
		pDevice->RxJumboDescCnt = 0;
	}

	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_StartTxDma(PLM_DEVICE_BLOCK pDevice, PLM_PACKET pPacket)
{
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_CompleteTxDma(PLM_DEVICE_BLOCK pDevice, PLM_PACKET pPacket)
{
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_IndicateRxPackets(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	struct sk_buff *skb;
	int size;

	while (1) {
		pPacket = (PLM_PACKET)
			QQ_PopHead(&pDevice->RxPacketReceivedQ.Container);
		if (pPacket == 0)
			break;
		if (pPacket->PacketStatus != LM_STATUS_SUCCESS) {
			QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
			continue;
		}
		pUmPacket = (PUM_PACKET) pPacket;
		pci_unmap_single(pUmDevice->pdev,
				pUmPacket->mapping,
				pPacket->u.Rx.RxBufferSize,
				PCI_DMA_FROMDEVICE);
		if ((size = pPacket->PacketSize) > pDevice->RxMtu) {
			continue;
		}
		skb = pUmPacket->skbuff;
		skb_put(skb, size);
		skb->pkt_type = 0;
		skb->protocol = eth_type_trans(skb, skb->dev);
		skb->ip_summed = CHECKSUM_NONE;
		if ((pDevice->ChipRevId >= T3_CHIP_REV_ID_B0) &&
			(skb->protocol == __constant_htons(ETH_P_IP))) {
			if (!(((struct iphdr *) skb->data)->frag_off &
				__constant_htons(IP_MF | IP_OFFSET))) {

				if (pPacket->Flags & LM_VALID_TCP_UDP_CHKSUM_FIELD)
					skb->ip_summed = CHECKSUM_UNNECESSARY;

			}
		}
#ifdef NICE_SUPPORT
		if( pUmDevice->nice_rx ) {
			vlan_tag_t *vlan_tag;

			vlan_tag = (vlan_tag_t *) &skb->cb[0];
			if (pPacket->Flags & LM_VALID_VLAN_TAG) {
				vlan_tag->signature = 0x7777;
				vlan_tag->tag = pPacket->VlanTag;
			}
			else {
				vlan_tag->signature = 0;
			}
			pUmDevice->nice_rx(skb, pUmDevice->nice_ctx);
		}
		else {
			netif_rx(skb);
		}
#else
		netif_rx(skb);
#endif

		skb = dev_alloc_skb(pPacket->u.Rx.RxBufferSize);
		if (skb == 0) {
			pUmPacket->skbuff = 0;
			QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, pPacket);
		}
		else {
			pUmPacket->skbuff = skb;
			skb->dev = pUmDevice->dev;
			skb_reserve(skb, 2);
			pUmPacket->mapping = pci_map_single(pUmDevice->pdev,
						skb->tail,
						pPacket->u.Rx.RxBufferSize,
						PCI_DMA_FROMDEVICE);
			pPacket->u.Rx.pRxBufferVirt = skb->tail;
			set_64bit_addr(&pPacket->u.Rx.RxBufferPhy, pUmPacket->mapping);
			QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
		}
	}
	return LM_STATUS_SUCCESS;
}

void
replenish_rx_buffers(PUM_DEVICE_BLOCK pUmDevice)
{
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	PLM_DEVICE_BLOCK pDevice = (PLM_DEVICE_BLOCK) pUmDevice;
	struct sk_buff *skb;

	while ((pUmPacket = (PUM_PACKET)
		QQ_PopHead(&pUmDevice->rx_out_of_buf_q.Container)) != 0) {
		pPacket = (PLM_PACKET) pUmPacket;
		if ((skb = dev_alloc_skb(pPacket->u.Rx.RxBufferSize)) == 0) {
           	QQ_PushTail(&pUmDevice->rx_out_of_buf_q.Container, pPacket);
			break;
		}
		pUmPacket->skbuff = skb;
		skb->dev = pUmDevice->dev;
		skb_reserve(skb, 2);
		pPacket->u.Rx.pRxBufferVirt = skb->tail;
		pUmPacket->mapping = pci_map_single(pUmDevice->pdev,
					skb->tail,
					pPacket->u.Rx.RxBufferSize,
					PCI_DMA_FROMDEVICE);
		set_64bit_addr(&pPacket->u.Rx.RxBufferPhy, pUmPacket->mapping);
		QQ_PushTail(&pDevice->RxPacketFreeQ.Container, pPacket);
	}
}

LM_STATUS
MM_IndicateTxPackets(PLM_DEVICE_BLOCK pDevice)
{
	PUM_DEVICE_BLOCK pUmDevice;
	PLM_PACKET pPacket;
	PUM_PACKET pUmPacket;
	struct sk_buff *skb;

	while (1) {
		pPacket = (PLM_PACKET)
			QQ_PopHead(&pDevice->TxPacketXmittedQ.Container);
		if (pPacket == 0)
			break;
		pUmPacket = (PUM_PACKET) pPacket;
		skb = pUmPacket->skbuff;
		pci_unmap_single(pUmDevice->pdev, pUmPacket->mapping, skb->len,
				 PCI_DMA_TODEVICE);
		dev_kfree_skb_irq(skb);
		pUmPacket->skbuff = 0;
		MM_ACQUIRE_TX_LOCK(pDevice, 0, FreeQ);
		QQ_PushTail(&pDevice->TxPacketFreeQ.Container, pPacket);
		MM_RELEASE_TX_LOCK(pDevice, 0, FreeQ);
	}
	pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	if (pUmDevice->tx_full) {
		if (QQ_GetEntryCnt(&pDevice->TxPacketFreeQ.Container) >=
			(QQ_GetSize(&pDevice->TxPacketFreeQ.Container) >> 1))

		pUmDevice->tx_full = 0;
		netif_wake_queue(pUmDevice->dev);
	}
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_IndicateStatus(PLM_DEVICE_BLOCK pDevice, LM_STATUS Status)
{
	PUM_DEVICE_BLOCK pUmDevice = (PUM_DEVICE_BLOCK) pDevice;
	struct net_device *dev = pUmDevice->dev;
	LM_FLOW_CONTROL flow_control;

	pUmDevice->delayed_link_ind = 0;
	if (Status == LM_STATUS_LINK_DOWN) {
		printk(KERN_INFO "%s: Link down\n", dev->name);
	}
	else if (Status == LM_STATUS_LINK_ACTIVE) {
		printk(KERN_INFO "%s: Link up, ", dev->name);
		if (pDevice->LineSpeed == LM_LINE_SPEED_1000MBPS)
			printk("1000 Mbps ");
		else if (pDevice->LineSpeed == LM_LINE_SPEED_100MBPS)
			printk("100 Mbps ");
		else if (pDevice->LineSpeed == LM_LINE_SPEED_10MBPS)
			printk("10 Mbps ");
		if (pDevice->DuplexMode == LM_DUPLEX_MODE_FULL)
			printk("full duplex");
		else
			printk("half duplex");

		flow_control = pDevice->FlowControl &
			(LM_FLOW_CONTROL_RECEIVE_PAUSE |
			LM_FLOW_CONTROL_TRANSMIT_PAUSE);
		if (flow_control) {
			if (flow_control & LM_FLOW_CONTROL_RECEIVE_PAUSE) {
				printk(", receive ");
				if (flow_control & LM_FLOW_CONTROL_TRANSMIT_PAUSE)
					printk("& transmit ");
			}
			else {
				printk(", transmit ");
			}
			printk("flow control ON");
		}
		printk("\n");
	}
	return LM_STATUS_SUCCESS;
}

LM_STATUS
MM_FreeRxBuffer(PLM_DEVICE_BLOCK pDevice, PLM_PACKET pPacket)
{
	PUM_PACKET pUmPacket;
	struct sk_buff *skb;

	pUmPacket = (PUM_PACKET) pPacket;
	skb = pUmPacket->skbuff;
	dev_free_skb(skb);
	pUmPacket->skbuff = 0;
	return LM_STATUS_SUCCESS;
}

#include "tigon3.c"


/*
 * Local variables:
 *  SMP-compile-command: "gcc -D__SMP__ -DMODULE -D__KERNEL__ -Wall -Wstrict-prototypes -O6 -c bcm5700.c `[ -f /usr/include/linux/modversions.h ] && echo -DMODVERSIONS`"
 *  compile-command: "gcc -DMODULE -D__KERNEL__ -Wall -Wstrict-prototypes -O6 -c bcm5700.c `[ -f /usr/include/linux/modversions.h ] && echo -DMODVERSIONS`"
 *  c-indent-level: 4
 *  c-basic-offset: 4
 *  tab-width: 4
 * End:
 */
