
/******************************************************************************/
/* The information contained in this file is confidential and proprietary to  */
/* Broadcom Corporation.  No part of this file may be reproduced or           */
/* distributed, in any form or by any means for any purpose, without the      */
/* express written permission of Broadcom Corporation.                        */
/*                                                                            */
/* (c) COPYRIGHT 2000 Broadcom Corporation, ALL RIGHTS RESERVED.              */
/*                                                                            */
/*                                                                            */
/* List Functions.                                                            */
/*      void       LL_InitList   (PLL_CONTAINER pList)                        */
/*      char       LL_Empty      (PLL_CONTAINER pList)                        */
/*      PLL_ENTRY  LL_GetHead    (PLL_CONTAINER pList)                        */
/*      PLL_ENTRY  LL_GetTail    (PLL_CONTAINER pList)                        */
/*      void       LL_PushHead   (PLL_CONTAINER pList, PLL_ENTRY pEntry)      */
/*      void       LL_PushTail   (PLL_CONTAINER pList, PLL_ENTRY pEntry)      */
/*      PLL_ENTRY  LL_PopHead    (PLL_CONTAINER pList)                        */
/*      PLL_ENTRY  LL_PopTail    (PLL_CONTAINER pList)                        */
/*      PLL_ENTRY  LL_NextEntry  (PLL_ENTRY pEntry)                           */
/*      PLL_ENTRY  LL_PrevEntry  (PLL_ENTRY pEntry)                           */
/*      void       LL_Unlink     (PLL_CONTAINER pList, PLL_ENTRY pEntry)      */
/*      char       LL_EntryInList(PLL_CONTAINER pList, PLL_ENTRY pEntry)      */
/*                                                                            */
/* History:                                                                   */
/*    02/25/00 Hav Khauv        Initial version.                              */
/******************************************************************************/

#ifndef LIST_H 
#define LIST_H



/******************************************************************************/
/* Doubly linked list structure used for linking entries together. */
/******************************************************************************/

typedef struct _LL_LIST_ENTRY {
    struct _LL_LIST_ENTRY *FLink;
    struct _LL_LIST_ENTRY *BLink;
} LL_LIST_ENTRY, *PLL_LIST_ENTRY;



/******************************************************************************/
/* List container */
/******************************************************************************/

typedef struct {
    LL_LIST_ENTRY   Link;
    unsigned long   EntryCount;
} LL_LIST_CONTAINER, *PLL_LIST_CONTAINER;



#ifdef LL_USE_MACROS
/* notdone */
#else

#ifdef LL_NO_INLINE
#define __inline
#endif /* LL_NO_INLINE */


/******************************************************************************/
/* Description:                                                               */
/*    Initializes a list container.                                           */
/*                                                                            */
/* Return:                                                                    */
/*    None.                                                                   */
/******************************************************************************/
__inline static void
LL_InitList(
PLL_LIST_CONTAINER pList) {
    pList->Link.FLink = pList->Link.BLink = (PLL_LIST_ENTRY) 0;
    pList->EntryCount = 0;
} /* LL_InitList */



/******************************************************************************/
/* Description:                                                               */
/*    Checks for an empty list.                                               */
/*                                                                            */
/* Return:                                                                    */
/*    TRUE if the list container is empty.                                    */
/******************************************************************************/
__inline static char 
LL_Empty(
PLL_LIST_CONTAINER pList) {
    return(pList->Link.FLink == (PLL_LIST_ENTRY) 0);
} /* LL_Empty */



/******************************************************************************/
/* Description:                                                               */
/*    Retrieves the head entry without removing the entry from the list.      */
/*                                                                            */
/* Return:                                                                    */
/*    Pointer to the head entry.                                              */
/******************************************************************************/
__inline static PLL_LIST_ENTRY 
LL_GetHead(
PLL_LIST_CONTAINER pList) {
    return pList->Link.FLink;
} /* LL_GetHead */



/******************************************************************************/
/* Description:                                                               */
/*    Retrieves the tail entry without removing it from the list.             */
/*                                                                            */
/* Return:                                                                    */
/*    Pointer to the tail entry.                                              */
/******************************************************************************/
__inline static PLL_LIST_ENTRY 
LL_GetTail(
PLL_LIST_CONTAINER pList) {
    return pList->Link.BLink;
} /* LL_GetTail */



/******************************************************************************/
/* Description:                                                               */
/*    Adds an entry to the head of the list.                                  */
/*                                                                            */
/* Return:                                                                    */
/*    None.                                                                   */
/******************************************************************************/
__inline static void
LL_PushHead(
PLL_LIST_CONTAINER pList, 
PLL_LIST_ENTRY pEntry) {
    pEntry->BLink = (PLL_LIST_ENTRY) 0;
    pEntry->FLink = pList->Link.FLink;

    if(pList->Link.FLink == (PLL_LIST_ENTRY) 0) {
        pList->Link.BLink = (PLL_LIST_ENTRY) pEntry;
    } else {
        pList->Link.FLink->BLink = pEntry;
    } /* if */

    pList->Link.FLink = (PLL_LIST_ENTRY) pEntry;

    pList->EntryCount++;
} /* LL_PushHead */



/******************************************************************************/
/* Description:                                                               */
/*    Adds an entry to the tail of the list.                                  */
/*                                                                            */
/* Return:                                                                    */
/*    None.                                                                   */
/******************************************************************************/
__inline static void
LL_PushTail(
PLL_LIST_CONTAINER pList, 
PLL_LIST_ENTRY pEntry) {
    pEntry->BLink = pList->Link.BLink;

    if(pList->Link.BLink) {
        pList->Link.BLink->FLink = pEntry;
    } else {
        pList->Link.FLink = pEntry;
    } /* if */

    pList->Link.BLink = pEntry;
    pEntry->FLink = (PLL_LIST_ENTRY) 0;

    pList->EntryCount++;
} /* LL_PushTail */



/******************************************************************************/
/* Description:                                                               */
/*    Removes an entry from the head of the list.                             */
/*                                                                            */
/* Return:                                                                    */
/*    The head entry.                                                         */
/******************************************************************************/
__inline static PLL_LIST_ENTRY 
LL_PopHead(
PLL_LIST_CONTAINER pList) {
    PLL_LIST_ENTRY pEntry;

    pEntry = pList->Link.FLink;

    if(pEntry) {
        pList->Link.FLink = pEntry->FLink;
        if(pList->Link.FLink == (PLL_LIST_ENTRY) 0) {
            pList->Link.BLink = (PLL_LIST_ENTRY) 0;
        } else {
            pList->Link.FLink->BLink = (PLL_LIST_ENTRY) 0;
        } /* if */

        pList->EntryCount--;
    } /* if(pEntry) */
    
    return pEntry;
} /* LL_PopHead */



/******************************************************************************/
/* Description:                                                               */
/*    Removes an entry from the tail of the list.                             */
/*                                                                            */
/* Return:                                                                    */
/*    Pointer to the tail entry.                                              */
/******************************************************************************/
__inline static PLL_LIST_ENTRY 
LL_PopTail(
PLL_LIST_CONTAINER pList) {
    PLL_LIST_ENTRY pEntry;

    pEntry = pList->Link.BLink;

    if(pEntry) {
        pList->Link.BLink = pEntry->BLink;
        if(pList->Link.BLink == (PLL_LIST_ENTRY) 0) {
            pList->Link.FLink = (PLL_LIST_ENTRY) 0;
        } else {
            pList->Link.BLink->FLink = (PLL_LIST_ENTRY) 0;
        }

        pList->EntryCount--;
    } /* if(pEntry) */
    
    return pEntry;
} /* LL_PopTail */



/******************************************************************************/
/* Description:                                                               */
/*    Advances to the next entry.                                             */
/*                                                                            */
/* Return:                                                                    */
/*    Pointer to the next entry.                                              */
/******************************************************************************/
__inline static PLL_LIST_ENTRY
LL_NextEntry(
PLL_LIST_ENTRY pEntry) {
    return pEntry->FLink;
} /* LL_NextEntry */



/******************************************************************************/
/* Description:                                                               */
/*    Advances to the previous entry.                                         */
/*                                                                            */
/* Return:                                                                    */
/*    Pointer to the previous entry.                                          */
/******************************************************************************/
__inline static PLL_LIST_ENTRY
LL_PrevEntry(
PLL_LIST_ENTRY pEntry) {
    return pEntry->BLink;
} /* LL_PrevEntry */



/******************************************************************************/
/* Description:                                                               */
/*    Gets the number of entries in the list.                                 */
/*                                                                            */
/* Return:                                                                    */
/*    The number of entry in the list.                                        */
/******************************************************************************/
__inline static unsigned long 
LL_GetSize(
PLL_LIST_CONTAINER pList) {
    return pList->EntryCount;
} /* LL_GetSize */



/******************************************************************************/
/* Description:                                                               */
/*    Removes an entry from the list.                                         */
/*                                                                            */
/* Return:                                                                    */
/*    None.                                                                   */
/******************************************************************************/
__inline static void
LL_Unlink(
PLL_LIST_CONTAINER pList,
PLL_LIST_ENTRY pEntry) {
    if(pList->Link.FLink != pEntry && pList->Link.BLink != pEntry) {
        pEntry->BLink->FLink = pEntry->FLink;
        pEntry->FLink->BLink = pEntry->BLink;
        pList->EntryCount--;
    } else if(pList->Link.FLink == pEntry && pList->Link.BLink != pEntry) {
        LL_PopHead(pList);
    } else {
        LL_PopTail(pList);
    } /* if */
} /* LL_Unlink */



/******************************************************************************/
/* Description:                                                               */
/*    Determines if the entry is in the list.                                 */
/*                                                                            */
/* Return:                                                                    */
/*    TRUE if the entry is in the list.                                       */
/******************************************************************************/
__inline static char
LL_EntryInList(
PLL_LIST_CONTAINER pList,
PLL_LIST_ENTRY pEntry) {
    PLL_LIST_ENTRY pTemp;

    pTemp = LL_GetHead(pList);
    while(pTemp) {
        if(pTemp == pEntry) {
            break;
        } /* if */

        pTemp = LL_NextEntry(pTemp);
    } /* while */

    return(pTemp == pEntry);
} /* LL_EntryInList */
#endif /* LL_USE_MACROS */



#endif /* LIST_H */
