
/******************************************************************************/
/*                                                                            */
/* Broadcom BCM5700 Linux Network Driver, Copyright (c) 2000 Broadcom         */
/* Corporation.                                                               */
/* All rights reserved.                                                       */
/*                                                                            */
/* This program is free software; you can redistribute it and/or modify       */
/* it under the terms of the GNU General Public License as published by       */
/* the Free Software Foundation, located in the file LICENSE.                 */
/*                                                                            */
/******************************************************************************/

#ifndef MM_H
#define MM_H

#include "lm.h"
#include "tigon3.h"

extern int MM_Packet_Desc_Size;

#define MM_PACKET_DESC_SIZE MM_Packet_Desc_Size

#define MAX_TX_Q 3
#define MAX_TX_LOCK MAX_TX_Q + 1
#define ReadyQ  0
#define ActiveQ 1
#define FreeQ   2

#if defined(__SMP__) && !defined(GLOBAL_LOCK)
#define MM_ACQUIRE_TX_LOCK(_pDevice, _Ring, _Queue) \
	while (test_and_set_bit(0, (void*)&((PUM_DEVICE_BLOCK)(_pDevice))->tx_lock[_Queue]) != 0);

#define MM_RELEASE_TX_LOCK(_pDevice, _Ring, _Queue) \
	clear_bit(0, (void*)&((PUM_DEVICE_BLOCK)(_pDevice))->tx_lock[_Queue]);
#define MM_ACQUIRE_SEND_BD_LEFT_LOCK_IN_DPC(_pDevice, _Ring) \
	while (test_and_set_bit(0, (void*)&((PUM_DEVICE_BLOCK)(_pDevice))->tx_lock[MAX_TX_LOCK - 1]) != 0);

#define MM_RELEASE_SEND_BD_LEFT_LOCK_IN_DPC(_pDevice, _Ring) \
	clear_bit(0, (void*)&((PUM_DEVICE_BLOCK)(_pDevice))->tx_lock[MAX_TX_LOCK - 1]);

#define MM_ACQUIRE_UNDI_LOCK(_pDevice) \
	spin_lock_irqsave(&_pDevice->undi_lock, _pDevice->undi_flags);

#define MM_RELEASE_UNDI_LOCK(_pDevice)
	spin_unlock_irqrestore(&_pDevice->undi_lock, _pDevice->undi_flags);
#else
#define MM_ACQUIRE_TX_LOCK(_pDevice, _Ring, _Queue)
#define MM_RELEASE_TX_LOCK(_pDevice, _Ring, _Queue)
#define MM_ACQUIRE_SEND_BD_LEFT_LOCK_IN_DPC(_pDevice, _Ring)
#define MM_RELEASE_SEND_BD_LEFT_LOCK_IN_DPC(_pDevice, _Ring)
#define MM_ACQUIRE_UNDI_LOCK(_pDevice)
#define MM_RELEASE_UNDI_LOCK(_pDevice)

#endif

#define MM_ACQUIRE_TX_LOCK_IN_DPC(_pDevice, _Ring, _Queue) \
	MM_ACQUIRE_TX_LOCK(_pDevice, _Ring, _Queue)

#define MM_RELEASE_TX_LOCK_IN_DPC(_pDevice, _Ring, _Queue) \
	MM_RELEASE_TX_LOCK(_pDevice, _Ring, _Queue)

#define MM_ACQUIRE_RX_LOCK(_pDevice)

#define MM_RELEASE_RX_LOCK(_pDevice)

#define MM_ACQUIRE_RX_LOCK_IN_DPC(_pDevice)

#define MM_RELEASE_RX_LOCK_IN_DPC(_pDevice)

#define MM_ACQUIRE_INT_LOCK_IN_DPC(_pDevice)
#define MM_RELEASE_INT_LOCK_IN_DPC(_pDevice)

#define MM_ACQUIRE_INT_LOCK(_pDevice)
#define MM_RELEASE_INT_LOCK(_pDevice)

#define MM_UINT_PTR(_ptr)   ((unsigned int) (_ptr))
#define MM_UINT32_PTR_LOW(_ptr) ((unsigned int) (_ptr))
#define MM_UINT32_PTR_HIGH(_ptr) ((unsigned int) 0)

#define DbgPrint(fmt, arg...) printk(KERN_WARNING fmt, ##arg)
#define DbgBreakPoint() __asm__("int $129")
#define MM_Wait(time) udelay(time)

#endif
