/*****************************************************************************
 *****************************************************************************

 Copyright (c) 1999 - 2001, Intel Corporation 

 All rights reserved.

 Redistribution and use in source and binary forms, with or without 
 modification, are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice, 
     this list of conditions and the following disclaimer.

  2. Redistributions in binary form must reproduce the above copyright notice,
     this list of conditions and the following disclaimer in the documentation 
     and/or other materials provided with the distribution.

  3. Neither the name of Intel Corporation nor the names of its contributors 
     may be used to endorse or promote products derived from this software 
     without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED. IN NO EVENT SHALL CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 

 *****************************************************************************
 *****************************************************************************/

#include "e1000.h"

void ResetAdaptive(PADAPTER_STRUCT Adapter, UINT32 LineSpeed)
{
    UINT32 TempReg;

    if (Adapter->AdaptiveIFS) {

        if (!Adapter->IFSParamsForced) {
            Adapter->CurrentIFSValue = 0;
            Adapter->IFSMinValue = MIN_IFS;
            Adapter->IFSMaxValue = MAX_IFS;
            Adapter->IFSStepSize = IFS_STEP;
            Adapter->IFSRatio = IFS_RATIO;
        }
        Adapter->InIFSMode = FALSE;
        TempReg = E1000_READ_REG(Ait);
        TempReg &= HIGH_16_BITS_MASK;
        E1000_WRITE_REG(Ait, TempReg);
    }

    if (Adapter->AdaptiveTxThreshold) {
        if (LineSpeed == 1000) {

            Adapter->CurrentTxThreshold = TX_THRESHOLD_DISABLE;
        } else {
            Adapter->CurrentTxThreshold = TX_THRESHOLD_START;
        }
        Adapter->TxThresholdTimer = 0;
        E1000_WRITE_REG(Ett, Adapter->CurrentTxThreshold);
    }
}

void UpdateAdaptiveIFS(PADAPTER_STRUCT Adapter,
                       UINT32 TxPackets, UINT32 TotalCollisions)
{
    UINT32 TempIFSReg;

    if (Adapter->AdaptiveIFS) {
        if ((TotalCollisions * Adapter->IFSRatio) > TxPackets) {

            if (TxPackets >= MIN_NUMBER_OF_TRANSMITS) {
                Adapter->InIFSMode = TRUE;
                if (Adapter->CurrentIFSValue < Adapter->IFSMaxValue) {
                    if (Adapter->CurrentIFSValue == 0)
                        Adapter->CurrentIFSValue = Adapter->IFSMinValue;
                    else
                        Adapter->CurrentIFSValue += Adapter->IFSStepSize;
                    TempIFSReg = E1000_READ_REG(Ait);
                    TempIFSReg &= HIGH_16_BITS_MASK;
                    TempIFSReg |= Adapter->CurrentIFSValue;
                    E1000_WRITE_REG(Ait, TempIFSReg);
                    DEBUGOUT1("IFS %d\n\r", Adapter->CurrentIFSValue);
                }
            }
        }

        else
            if ((Adapter->InIFSMode == TRUE)
                && (TxPackets <= MIN_NUMBER_OF_TRANSMITS)) {
            Adapter->CurrentIFSValue = 0;
            Adapter->InIFSMode = FALSE;
            DEBUGOUT1("IFS %d\n\r", Adapter->CurrentIFSValue);
            TempIFSReg = E1000_READ_REG(Ait);
            TempIFSReg &= HIGH_16_BITS_MASK;
            TempIFSReg |= Adapter->CurrentIFSValue;
            E1000_WRITE_REG(Ait, TempIFSReg);
        }
    }
}

BOOLEAN UpdateAdaptiveTxThreshold(PADAPTER_STRUCT Adapter,
                                  UINT32 TxPackets,
                                  UINT32 TxUnderruns,
                                  UINT32 CallbackInterval)
{
    UINT32 ReturnValue = TRUE;

    if (Adapter->AdaptiveTxThreshold &&
        (Adapter->CurrentTxThreshold != TX_THRESHOLD_DISABLE)) {

        if ((TxUnderruns == 0) &&
            (TxPackets != 0) &&
            (Adapter->CurrentTxThreshold > TX_THRESHOLD_START)) {
            Adapter->TxThresholdTimer++;
            if ((Adapter->TxThresholdTimer * CallbackInterval) ==
                TX_THRESHOLD_TIMER_MS) {
                Adapter->CurrentTxThreshold -= TX_THRESHOLD_DECREMENT;
                E1000_WRITE_REG(Ett, Adapter->CurrentTxThreshold);
                Adapter->TxThresholdTimer = 0;
            }
        }

        if (TxUnderruns != 0) {
            Adapter->TxThresholdTimer = 0;
            Adapter->CurrentTxThreshold += TX_THRESHOLD_INCREMENT;
            if (Adapter->CurrentTxThreshold > TX_THRESHOLD_STOP) {
                Adapter->CurrentTxThreshold = TX_THRESHOLD_DISABLE;
                E1000_WRITE_REG(Ett, 0);
                ReturnValue = FALSE;
            } else {
                E1000_WRITE_REG(Ett, Adapter->CurrentTxThreshold);
            }
        }
    }
    return ReturnValue;
}
